//  This is a small identd (RFC 1413) server; it returns the USER
//  environment variable in response to any well-formed identd request.
//  It has the following options:
//
//  usage: identd [-h] [[-p port] [-t timeout] [-c #connect] [-u user]]
//
//    -h       Display usage.
//
//    -p n,   where n is the port to listen to.  (Default: 113)
//
//    -t n,   where n is how many seconds to wait between a connection and
//            the completion of the transaction.  (Default: 30)
//
//    -c n,   number of connects allowed on socket. (Default: 10)
//
//    -u user response string, overrides environment variable USER (Default: os2user)
//
//   accepts rfc1413 (identd) requests, and says either "ERROR: INVALID-PORT"
//   if it couldn't understand the request, or "OS/2: FOOBAR" where
//   'foobar' is whatever your USER environment variable is set to (or 'os2-user'
//   if it's not set or the option passed by -u)
//
//   Original: 1.1 Roddy's wee identd/2  - 31 October 1997
//
//   Version 1.2 Michael Greene <greenemk@cox.net> - 08 August 2004
//
//   Compiles with Watcom 1.3
//
#define INCL_DOSPROCESS
#define INCL_ERRORS
#define INCL_DOSERRORS
#define INCL_DOSSEMAPHORES

#include <os2.h>
#include <iostream>
using namespace std;
#include <cstring>
#include <cstdio>
#include <cstdlib>
//#include <cctype>
#include <types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa\inet.h>
#include <netdb.h>
#include <sys\un.h>
#include <unistd.h>
#include <process.h>
#include <time.h>
#include "identd.hpp"

int  setup_server(int port);
void message(int);
void cleanup(void);
void get_time(void);
void handle_error(int msgnum);
int  accept_connection(int sock);
void handle_connection(int sock, int timeout);
void transact_ident(void *arg);

static int errorcode  = 0;
static int serverSock = 0;

OPTIONS prg_options = { DEFAULT_PORT,
                        DEFAULT_TIMEOUT,
                        DEFAULT_VERBOSITY,
                        DEFAULT_CONNECTS
                      };

int main(int argc, char *argv[])
{
// Get the USER response variable.  If USER is set then use it or set
// it to the DEFAULT_RESPONSE_STRING which is "os2user" defined in
// identd.hpp.  Can over ride with the command line option [-u user]
   char defrespvar[]=DEFAULT_RESPONSE_VARIABLE;
   if (getenv(defrespvar) == 0) prg_options.respvar = DEFAULT_RESPONSE_STRING;
   else prg_options.respvar = getenv(defrespvar);

// Parse the command line arguments and assign, if required.
//
//  identd [-h] [-c connectlimit] [-p port] [-t connecttimeout] [-u user] [-v verbosity]
//
   char * arg_process;
   if (argc == 2)  // 2 args must be the help option but check anyway
   {
      arg_process = argv[1];
      if ((arg_process[0]=='-')&((arg_process[1] == 'h')|(arg_process[1]=='H')))
                                             handle_error(1);  // Errorcode 1 - exit
      handle_error(2);  // Errorcode 2 - exit
   }
   if (argc != 1)  // No options - run default
   {
      if (((argc-1)%2 != 0) | (argc > 11)) handle_error(2);  // Errorcode 2 - exit
      for (int assignargs = 1; assignargs < argc; assignargs = assignargs + 2)
      {
         arg_process = argv[assignargs];
         if (arg_process[0] == '-')
         {
            switch (arg_process[1])
            {
               case 'c' : prg_options.connects = atoi(argv[assignargs+1]);
                          break;
               case 'h' : handle_error(1);  // Errorcode 1 - exit
                          break;
               case 'p' : prg_options.useport = atoi(argv[assignargs+1]);
                          break;
               case 't' : prg_options.timeout = atoi(argv[assignargs+1]);
                          break;
               case 'u' : prg_options.respvar = argv[assignargs+1];
                          break;
               case 'v' : prg_options.verbose = atoi(argv[assignargs+1]);
                          if (prg_options.verbose < 0 || prg_options.verbose > 2)
                                             handle_error(5);  // Errorcode 5 - exit
                          break;
               default  : handle_error(4);  // Errorcode 4 - exit
                          break;
            }
         }
         else handle_error(3);  // Errorcode 3 - exit
      }
   }
   // end of parse args - open the socket and wait.

   cout << "Starting with options:" << endl;
   cout << "Port: " << prg_options.useport
        << "  Timeout: " << prg_options.timeout
        << "  User: " << prg_options.respvar
        << "  Connects: " << prg_options.connects
        << endl << endl;

   serverSock = setup_server(prg_options.useport);

   // EXLST_ADD (1) Add an address to the termination list.
   // Org: DosExitList(EXLST_ADD | 0x00000000, (PFNEXITLIST) cleanup);
   DosExitList(EXLST_ADD | 0x00000000, (PFNEXITLIST) cleanup);

   int  clientSock;
   while (1)
   {
      clientSock = accept_connection(serverSock);
      handle_connection(clientSock, prg_options.timeout);
      if (shutdown(clientSock,2) == -1) message(15);  // Errorcode 15
      if (soclose(clientSock) == -1) handle_error(15);  // Errorcode 15 - exit
   }
}


// cleanup()
//
//  Added to ExitList (DosExitList) to cleanup on exit.
//
void cleanup(void)
{
   cout << "\nExiting errorcode " << errorcode << endl;
   if (soclose(serverSock) == -1) message(16);  // Errorcode 16
}


// handle_error
//
//  Display messages and exit.
//
void handle_error(int msgnum)
{
   message(msgnum);
   exit(msgnum);
}


// get_time
//
void get_time(void)
{
   time_t t;
   char * ptr;
   t = time(NULL);
   time(&t);
   ptr = strtok(ctime(&t), "\n");
   cout << ptr;
}


// messages(n) - display version or error messages and then
//                usage.
//
//  Where n:
//
//  1     - Version string.
//  2     - Wrong number of arguments !!NOT USED!!.
//  3     - Bad arguments (prefix "-" not found).
//  4     - Bad arguments (found "-" but not p,t,v).
//  5     - Bad arguments (verbose < 1 or > 3).
//  6-9   - Error in socket open function.
//  10    - Error in accept() call! Print error stderr.
//  11-14 - Error in handle_connection (Sem / thread error).
//  15    - Error client socket close error
//  16    - Error server socket close error
//
void message(int msgdisplay)
{
   cout << endl;
   switch (msgdisplay)
      {
         case  1 : cout << "IDENTD v1.2 for OS/2 | 08 AUG 2004 Michael Greene"
                        << endl << "Compiled with Watcom 1.3rc3" << endl;break;
         case  2 : cout << "Error: Wrong number of arguments passed.";break;
         case  3 : cout << "Error: Bad argument(s), no - prefix.";break;
         case  4 : cout << "Error: Bad argument(s), could not parse.";break;
         case  5 : cout << "Error: Bad verbose argument passed.";break;
         case  6 : cout << "\nError: Bad reset sock_init().";break;
         case  7 : cout << "\nError: Bad socket socket().";break;
         case  8 : cout << "\nError: Bad socket bind().";break;
         case  9 : cout << "\nError: Bad socket listen().";break;
         case 10 : cout << "\nError: Bad socket accept().\n";break;
         case 11 : cout << "\nError: couldn't create a semaphore.\n";break;
         case 12 : cout << "\nError: couldn't create a thread.\n";break;
         case 13 : cout << "\nError: thread / socket timeout.\n";break;
         case 14 : cout << "\nError: unexpected error waiting for thread / socket.\n";break;
         case 15 : cout << "\nError: client socket close error.\n";break;
         case 16 : cout << "\nError: server socket close error.\n";break;
         case 17 : cout << "\nError: receive error.\n";break;
         case 18 : cout << "\nError: send error.\n";break;
         case 19 : cout << "\nError: DosPostEventSem error.\n";break;
      }

   if (msgdisplay < 6)
   {
      cout << "\n\n"
              "usage: identd [-h] [[-p port] [-t timeout] [-c #connect] [-u user]]\n\n"
              " -h       Display usage.\n\n"
              " -c n,    Number of connects allowed on socket. (Default: 10)\n\n"
              " -t n,    where n is how many seconds to wait between a connection\n"
              "          and the completion of the transaction.  (Default: 30)\n\n"
              " -p n,    where n is the port to listen to.  (Default: 113)\n\n"
              " -u user  Response string, overrides environment variable USER.\n"
              "          (Default: os2user)\n\n";
   }
}


// setup_server
//
//  <NETINET\IN.H>:
//   struct in_addr {u_long s_addr;};
//   struct sockaddr_in
//      {
//      u_char  sin_len;
//      u_char  sin_family;
//      u_short sin_port;
//      struct  in_addr sin_addr;
//      char    sin_zero[8];
//      };
//
//  1. init sock_init() - initialize the socket data structures and check
//     to see if INET.SYS is running. (1 - Failure 0 - Success)
//
//  2. socket()
//       #include <types.h>
//       #include <sys\socket.h>
//       int socket(int domain, int type, int protocol)
//     return -1 error, > 0 success
//
//  3. mmset()
//       #include <string.h> [cstring]
//       void *memset( void *dst, int c, size_t length );
//     The memset function fills the first length characters of the object
//     pointed to by dst with the value c.
//
//  4. bind()
//       #include <types.h>
//       #include <sys\socket.h>
//       int bind(int s, struct sockaddr *name, int namelen)
//     int s returned by socket()
//
//  5. listen()
//       #include <types.h>
//       #include <sys\socket.h>
//       #include <netinet\in.h>
//       int listen(int s, int backlog)
//     int s returned by socket()
//
//  *note: listen() set to reject after 5 connects.  sys/socket.h allows
//         from 0 to SOMAXCONN (1024).
//
int setup_server(int port)
{
   struct sockaddr_in serv_addr;
   int sockdescript;
   BOOL flag = 1;

   if (sock_init()) handle_error(6);  // Errorcode 6 - exit
   sockdescript = socket(AF_INET, SOCK_STREAM, 0);

   if (sockdescript == -1) handle_error(7);  // Errorcode 7 - exit;

   memset(&serv_addr, (char) 0, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   serv_addr.sin_port = htons(port);

   // setsockopt so I can stop and restart the server - set socket to reuse.
   if (setsockopt(sockdescript, SOL_SOCKET, SO_REUSEPORT,(const char*)&flag,sizeof(flag)) == -1) cout << "setsockopt error";

   if (bind(sockdescript, (struct sockaddr *) &serv_addr, sizeof(serv_addr))) handle_error(8);  // Errorcode 8 - exit

   if (listen(sockdescript, prg_options.connects)) handle_error(9);  // Errorcode 9 - exit

   return(sockdescript);
}


// accept_connection
//
//
//  <NETINET\IN.H>:
//   struct in_addr {u_long s_addr;};
//   struct sockaddr_in
//      {
//      u_char  sin_len;
//      u_char  sin_family;
//      u_short sin_port;
//      struct  in_addr sin_addr;
//      char    sin_zero[8];
//      };
//
//  1. accept()
//       #include <types.h>
//       #include <sys\socket.h>
//       #include <netinet\in.h>
//       #include <netnb\nb.h>
//       #include <sys\un.h>
//     int accept(int s, sockaddr *name, int *namelen)
//     This call is used by a server acting in a connection-oriented mode to accept
//     a connection request from a client.  The call accepts the first connection on
//     its queue of pending connection requests. The accept() call creates a new
//     socket descriptor with the same properties as s and returns it to the caller.
//     The new socket descriptor cannot be used to accept new connections. The
//     original socket, s, remains available to accept more connection requests.
//
int accept_connection(int sock)
{
   int newSock;
   int addrSize;

   struct sockaddr_in clientAddr;

   addrSize = sizeof(clientAddr);

   newSock = accept(sock, (struct sockaddr *)&clientAddr, &addrSize);
   if (newSock == -1) handle_error(10);  // Errorcode 10 - exit;

   get_time();

   struct hostent *h = gethostbyaddr((char *) &clientAddr.sin_addr,
                       sizeof(clientAddr.sin_addr),
                       AF_INET);

   char *hname = (h == 0) ? (char *) inet_ntoa(clientAddr.sin_addr) : h->h_name;

   cout << " : connect from "
        << (char *) inet_ntoa(clientAddr.sin_addr)
        << " [" << hname << " ]\n";

   return(newSock);
}


// handle_connection
//
//  1. Create tp (THREAD_PARMS structure defined in identd.hpp) using
//     new THREAD_PARAMS memory size of THREAD_PARAMS
//
//       struct THREAD_PARAMS
//       {
//          int sock;
//          int timeout;
//          HEV hevActivity;
//       };
//
//  2. Create an event semaphore (DosCreateEventSem) - unnamed.
//
//  2. _beginthread
//
//  3. Wait for complete or timeout (DosWaitEventSem)
//
//  4. DosCloseEventSem and get out.
//
void handle_connection(int sock, int timeout)
{
   int tid;
   int rc;

   THREAD_PARAMS * tp = new THREAD_PARAMS;
                tp->sock    = sock;
                tp->timeout = timeout;

   if (DosCreateEventSem(NULL, &tp->hevActivity, 0L, FALSE)) errorcode = 11;
   else
   {
      tid = _beginthread(transact_ident, NULL, 16*1024, tp);
      if (tid == -1) errorcode = 12;
      else
      {
         rc = DosWaitEventSem(tp->hevActivity, timeout * 1000);
         switch (rc)
         {
            case NO_ERROR:      break;
            case ERROR_TIMEOUT: get_time();
                                cout << " : ERROR : CONNECTION TIMEOUT\n";
                                if (DosKillThread(tid)) cout << "Timeout: Killing thread failed\n";
                                break;
            default:            if (DosKillThread(tid)) cout << "Killing thread in default failed\n";
                                break;
         }
      }
      DosCloseEventSem(tp->hevActivity);
   }
   delete tp;
}


// transact_ident
//
//  1. recv()
//       #include <types.h>
//       #include <sys\socket.h>
//       int recv(int s, char *buf, int len, int flags)
//
//  2. send()
//       #include <types.h>
//       #include <sys\socket.h>
//       int send(int s, char *msg, int len, int flags)
//
//  3. DosPostEventSem()
//       #define INCL_DOSSEMAPHORES
//       #include <os2.h>
//       int = DosPostEventSem(hev)
//
//  4. endthread
//
void transact_ident(void *arg)
{
   THREAD_PARAMS *tp = (THREAD_PARAMS *) arg;

   int port_server;
   int port_client;
   int  bytes_received;
   char working_buffer[BUFFER_SIZE];

   bytes_received = recv(tp->sock, working_buffer, BUFFER_SIZE, 0);

   if (bytes_received == -1)  handle_error(17);  // recv() error
   if (bytes_received ==  0)
   {
      cout << endl;
      get_time();
      cout << "0, 0, UNKNOWN-ERROR"; // 0 bytes, can this happen? Doubt it.
   }

   get_time();
   cout << " : Request was " << working_buffer;

   if (bytes_received ==  1) cout << endl;  // 1 bytes, catch a telnet connect
   if (sscanf(working_buffer, "%d , %d", &port_server, &port_client) != 2)
   {
      sprintf(working_buffer, "0 , 0 : ERROR : INVALID-PORT");
   }
   else
   {
      sprintf(working_buffer, "%d, %d : OS/2 : %s", port_server,
                                                    port_client,
                                                    prg_options.respvar);
   }

   get_time();
   cout << " : Response was " << working_buffer << endl;

   // send response
   if (send(tp->sock, working_buffer, strlen(working_buffer), 0) == -1)
                                                  handle_error(18);  // send() error
   sprintf(working_buffer, "\r\n");

   if (DosPostEventSem(tp->hevActivity)) handle_error(19);  // error;
   _endthread();
}

